/**
  *  \file client/dialogs/baseshipbuild.cc
  *  \brief WBaseShipBuildDialog dialog
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/baseshipbuild.h"
#include "client/tasksupport.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "game/actions/baseship.h"
#include "game/error.h"
#include "int/exec.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/widgets/button.h"
#include "ui/widgets/statictext.h"
#include "util/nls.h"
#include "game/global.h"
#include "ui/commands.h"
#include "game/ship.h"
#include "game/fcode.h"
#include "game/planettransfer.h"

static const TBuildOrder blankBuildOrder = { 0, 0, 0, 0, 0, 0, 0 };

/** Constructor.
    \param planet Planet to work on
    \param hulls Hulls to offer for building (will be cleared) */
WBaseShipBuildDialog::WBaseShipBuildDialog(GPlanet& planet, std::vector<int>& hulls)
    : WShipBuildDialog(_("Build a Ship"), planet, true, hulls),
      planet(planet)
{
    init();
}

void
WBaseShipBuildDialog::init()
{
    // UIGroup[UIHBoxLayout]
    //   UIButton "U"
    //   UIStaticText "Use parts from storage"
    // UIGroup[UIHBoxLayout]
    //   UIButton "Enter"
    //   UIButton "ESC"
    //   [UIButton "Cancel"]
    //   UISpacer
    //   UIButton "Help"
    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));

    UIButton& ok = h.add(new UIButton(2000, _("Enter - Build"), SDLK_RETURN));
    UIButton& cancel = h.add(new UIButton(2001, _("Exit"), SDLK_ESCAPE));
    UIButton& bill = h.add(new UIButton(2004, _("D - Detailed Bill"), 'd'));
    g.add(ok);
    g.add(cancel);
    g.add(bill);
    if (planet.getBaseBuildHull()) {
        UIButton& cancel_build = h.add(new UIButton(2002, UIKeyString(_("C - Cancel Build"))));
        cancel_build.sig_fired.add(this, &WBaseShipBuildDialog::onCancelBuild);
        g.add(cancel_build);
    }
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(_("Help"), 'h')));

    UIButton& u = h.add(new UIButton(2003, "U", 'u'));
    g2.add(u);
    g2.add(h.add(new UIStaticText(_("Use parts from storage"), UIStaticText::WhiteTitle)));
    g2.add(h.add(new UISpacer()));

    add(g2);
    add(g);
    add(h.add(new WHelpWidget("pcc2:buildship")));

    ok.sig_fired.add(this, &WBaseShipBuildDialog::onOk);
    cancel.sig_fired.add(this, &WBaseShipBuildDialog::onCancel);
    u.sig_fired.add(this, &WBaseShipBuildDialog::onToggleUseParts);
    bill.sig_fired.add(static_cast<WShipBuildDialog*>(this), &WShipBuildDialog::onDetailedBill);

    updateUseParts();
    pack();
}

WBaseShipBuildDialog::~WBaseShipBuildDialog()
{ }

void
WBaseShipBuildDialog::onOk()
{
    GPlanetTransfer transfer(planet);
    GStarbaseBuildTransaction ta(action, transfer);
    if (GError e = ta.getError()) {
        if (e.isError(GError::eNoResource)) {
            bool add =
                UIMessageBox(_("You do not have enough resources to build this ship now. "
                               "Do you want to add this build order to this base's Auto Task, "
                               "to build it as soon as resources are available?"),
                             _("Build order rejected")).doYesNoDialog();
            ta.doCancel();
            transfer.doCancel();
            if (add && addToAutoTask(getAutoTaskForObject(planet, IntExecutionContext::pkBaseTask, true),
                                     makeBuildOrderCommand("EnqueueShip", action.getBuildOrder())))
            {
                action.getPlanet().setBaseBuildOrder(blankBuildOrder);
                stop(1);
            }
        } else {
            messageBox(format(_("You can not build this ship: %s"), e.getUserError()),
                       _("Build order rejected"));
            ta.doCancel();
            transfer.doCancel();
        }
    } else {
        if (int sid = planet.findShipCloningHere(getDisplayedTurn().getCurrentUniverse())) {
            UIMessageBox mbx((format(_("This base is already cloning %s. Do you want to cancel that order? "
                                       "If you say \"No\", this ship will be built after the clone completed."),
                                     getDisplayedTurn().getCurrentUniverse().getShip(sid).getName(GObject::PlainName))),
                             _("Build Ship"));
            UIKeyString y(_("Yes")), n(_("No"));
            mbx.addButton(cm_Enter,  y.getString(), y.getKey());
            mbx.addButton(1000,      n.getString(), n.getKey());
            mbx.addButton(cm_Escape, _("Cancel"),   SDLK_ESCAPE);
            mbx.addKey(cm_Enter, ' ');
            int reply = mbx.run();
            if (reply == cm_Escape) {
                return;
            }
            if (reply == cm_Enter) {
                cancelAllCloneOrders(planet);
            }
        }
        if (action.isChange()) {
            if (!UIMessageBox(_("This starbase already has a build order. Do you want to change that order?\n"
                               "To exit the ship build screen without changing the build order, use \"Exit\" (ESC)."),
                              _("Build Ship")).doYesNoDialog()) {
                //ta.doCancel();
                //transfer.doCancel();
                return;
            }
        }
        ta.doCommit();
        transfer.doCommit();
        action.getPlanet().setBaseBuildOrder(action.getBuildOrder());
        stop(1);
    }
}

void
WBaseShipBuildDialog::onCancel()
{
    stop(0);
}

void
WBaseShipBuildDialog::onCancelBuild()
{
    action.getPlanet().setBaseBuildOrder(blankBuildOrder);
    stop(1);
}

void
WBaseShipBuildDialog::onToggleUseParts()
{
    action.setUsePartsFromStorage(!action.isUsePartsFromStorage());
    updateUseParts();
    onPlanetChanged(); // FIXME: should use signals properly.
}

void
WBaseShipBuildDialog::updateUseParts()
{
    setStateOf(2003, UIButton::bf_Highlight, action.isUsePartsFromStorage());
}

void
cancelAllCloneOrders(GPlanet& planet)
{
    GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
    for (int sid = 1; sid <= NUM_SHIPS; ++sid) {
        if (univ.getShip(sid).isCloningAt(planet)) {
            univ.getShip(sid).setFCode(GFCode::generateRandomFCode());
        }
    }
}
