/**
  *  \file client/dialogs/navchart.cc
  *  \brief Ship Navigation Chart
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/navchart.h"
#include "client/dlg-visual.h"
#include "client/screens.h"
#include "client/widgets/helpwidget.h"
#include "client/wscanchart.h"
#include "cpluslib/format.h"
#include "game/formula.h"
#include "game/host.h"
#include "game/objlist.h"
#include "game/ship.h"
#include "game/univ.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/unicodechars.h"
#include "game/fleet.h"
#include "game/config.h"
#include "client/widgets/objlist.h"
#include "ui/commands.h"
#include "client/widgets/scanresult.h"

namespace {
    /** Navigation chart widget. */
    class WShipNavigationChart : public WScannerChartWidget {
        /* Position history */
        enum { MAX_OLD_POS = 20 };
        GPoint old_pos[MAX_OLD_POS];
        int old_pos_head;
        int old_pos_count;

        /* Chunnel operation */
        bool chunnel_mode;
        UIAbstractButton* chunnel_button;
        const GShip* chunnel_ship;

        /* Warp circles */
        bool circle_hyp;
        int circle_radius;

        /* Ship */
        GUniverse& univ;

        /* Alternative warp circles */
        GPoint alt_circles;

        void drawCircles(GfxCanvas& can, GPoint pt);

     public:
        WShipNavigationChart(GUniverse& univ);

        void setCircleRadius(bool hyp, int r);
        void setChunnelShip(GShip& ship);
        
        void drawPre(GfxCanvas& can);
        bool handleEvent(const UIEvent& event, bool second_pass);
        void setChunnelButton(UIAbstractButton* btn);
        void toggleChunnel();
        bool isChunnelMode() const;
        void doItemLock(GPoint clicked, TLockMode mode, bool marked_only, bool optimize_warp);
        void onDblClick();
    };
}

/** Check for valid chunnel mate.
    \param initiator Chunnel initiator
    \param mate Potential chunnel mate */
static bool
isValidChunnelMate(const GShip& initiator, const GShip& mate)
{
    int32_t dist = host.isPHost() ? square(config.MinimumChunnelDistance()) : 9901;
    return mate.isPlayable(GObject::Playable)
        && mate.getOwner() == initiator.getOwner()
        && mate.getFleetNumber() == 0
        && (mate.canDoSpecial(hf_FirecloudChunnel) || mate.canDoSpecial(hf_ChunnelTarget))
        && (distanceSquared(initiator.getPos(), mate.getPos()) >= dist);
}

/** Choose chunnel mate from position.
    \param initiator Initiator of chunnel
    \param pt Location to check
    \return ship Id, 0 if none found or canceled */
static int
chooseChunnelMate(GUniverse& univ, const GShip& initiator, GPoint pt)
{
    GObjectList objs;
    for (int i = univ.ty_played_ships.findNextIndex(0); i != 0; i = univ.ty_played_ships.findNextIndex(i)) {
        GShip& sh = univ.ty_played_ships.getObjectByIndex(i);
        if (sh.getPos() == pt && isValidChunnelMate(initiator, sh)) {
            objs.addObject(GObjectReference(univ.ty_played_ships, i));
        }
    }

    if (objs.size() == 0) {
        messageBox(_("There are no potential chunnel mates at the current position."), _("Chunnel"));
        return 0;
    } else if (objs.size() == 1) {
        return objs[0]->getId();
    } else {
        GObjectReference id = WObjectList(0, 10, 20*font_ems[FONT_NORMAL], objs).doStandardDialog2(_("Chunnel"), "pcc2:navchart");
        if (id.isValid()) {
            return id->getId();
        } else {
            return 0;
        }
    }
}

/** Constructor. */
WShipNavigationChart::WShipNavigationChart(GUniverse& univ)
    : old_pos_head(0),
      old_pos_count(0),
      chunnel_mode(false),
      chunnel_button(0),
      chunnel_ship(0),
      circle_hyp(false),
      circle_radius(0),
      univ(univ),
      alt_circles(0, 0)
{ }

/** Draw warp circles at given position.
    \param can Canvas to operate on
    \param pt Game universe location */
void
WShipNavigationChart::drawCircles(GfxCanvas& can, GPoint pt)
{
    GChartViewport& vp = getViewport();
    GfxPoint p = vp.scaleXY(pt);
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.setColor(COLOR_DARK);
    if (circle_hyp) {
        // xxxship
        // if (ship.isHyperdriving()) {
        // HYP circle
        drawCircle(ctx, p, vp.scale(340));
        drawCircle(ctx, p, vp.scale(360));
    } else if (circle_radius) {
        // ship.getWarp() >= 5) {
        // Warp circles
        drawCircle(ctx, p, vp.scale(circle_radius));
        drawCircle(ctx, p, vp.scale(2*circle_radius));
        if (vp.divi >= 2*vp.mult) {
            drawCircle(ctx, p, vp.scale(3*circle_radius));
        }
    } else {
        // no circles
    }
}

void
WShipNavigationChart::setCircleRadius(bool hyp, int r)
{
    circle_hyp = hyp;
    circle_radius = r;
}

void
WShipNavigationChart::setChunnelShip(GShip& ship)
{
    chunnel_ship = &ship;
}

void
WShipNavigationChart::drawPre(GfxCanvas& can)
{
    if (chunnel_mode) {
        /* Chunnel mode */
        GChartViewport& vp = getViewport();
        GfxContext ctx;
        ctx.useCanvas(can);
        ctx.useColorScheme(GfxStandardColorScheme::instance);

        /* Forbidden area */
        int dist = host.isPHost() ? config.MinimumChunnelDistance() : 100;
        if (dist > 0) {
            GfxPoint pt = vp.scaleXY(getScannerOrigin());
            int r = vp.scale(dist);
            ctx.setFillPattern(LTSLASH_FILL);
            ctx.setColor(COLOR_FIRE+4);
            drawFilledCircle(ctx, pt, r);
            drawCircle(ctx, pt, r);
        }

        /* Possible targets */
        ctx.setColor(COLOR_BRIGHTCYAN);  /* PCC 1.x uses GREENSCALE+15 */
        int r = vp.scale(10);
        if (r < 5)
            r = 5;
        if (r > 20)
            r = 20;
        for (int i = univ.ty_any_ships.findNextIndex(0); i != 0; i = univ.ty_any_ships.findNextIndex(i)) {
            const GShip& mate = univ.getShip(i);
            if (isValidChunnelMate(*chunnel_ship, mate)) {
                GfxPoint pt = vp.scaleXY(mate.getPos());
                ctx.setCursor(GfxPoint(pt.x - r, pt.y));
                drawLineRel(ctx, r, -r);
                drawLineRel(ctx, r, r);
                drawLineRel(ctx, -r, r);
                drawLineRel(ctx, -r, -r);
            }
        }
    } else {
        /* Regular mode, warp circles */
        drawCircles(can, getScannerOrigin());
        if (alt_circles.x > 0) {
            drawCircles(can, alt_circles);
        }
    }
}

bool
WShipNavigationChart::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
         case SDLK_TAB:
            // Push position
            old_pos[old_pos_head] = getCenterPosition();
            old_pos_head = (old_pos_head + 1) % MAX_OLD_POS;
            if (old_pos_count < MAX_OLD_POS)
                ++old_pos_count;

            // Go there
            setCenterPosition(getScannerTarget());
            return true;

         case SDLK_BACKSPACE:
            if (old_pos_count > 0) {
                // Pop position
                old_pos_head = (old_pos_head + MAX_OLD_POS-1) % MAX_OLD_POS;
                --old_pos_count;
                setCenterPosition(old_pos[old_pos_head]);
            }
            return true;

         case SDLK_INSERT:
         case 'y':
            // Add circle
            if (!chunnel_mode) {
                alt_circles = getScannerTarget();
                drawWidget();
            }
            return true;

         case SDLK_DELETE:
            // Remove circle
            if (!chunnel_mode) {
                alt_circles = GPoint(0, 0);
                drawWidget();
            }
            return true;

         case 'l':
            // List ships
            doListShips(getScannerTarget(), ls_IncludeForeignShips | ls_AllowForeignShips, 0, _("List Ships"), _("OK"));
            return true;

         case SDLK_F5:
         case ss_Ctrl + SDLK_F5:
            // Planet info
            if (int pid = univ.getPlanetAt(getScannerTarget(), true)) {
                doPlanetScan(pid);
            }
            return true;
        }
    }
    return WScannerChartWidget::handleEvent(event, second_pass);
}

void
WShipNavigationChart::setChunnelButton(UIAbstractButton* btn)
{
    chunnel_button = btn;
    if (chunnel_button != 0) {
        chunnel_button->setState(UIAbstractButton::bf_Highlight, chunnel_mode);
    }
}

void
WShipNavigationChart::toggleChunnel()
{
    if (chunnel_ship != 0) {
        chunnel_mode = !chunnel_mode;
        setChunnelButton(chunnel_button);    // updates the button's status
        drawWidget();
    }
}

bool
WShipNavigationChart::isChunnelMode() const
{
    return chunnel_mode;
}

/** Perform lock on target. Called whenever the user presses
    Space/Return or a mouse button to lock onto an object. This
    overrides the behaviour to lock only on Fireclouds in chunnel
    mode. */
void
WShipNavigationChart::doItemLock(GPoint clicked, TLockMode mode, bool marked_only, bool optimize_warp)
{
    if (chunnel_mode) {
        const GChartViewport& vp = getViewport();
        GLockData locker(clicked,
                         GPoint(vp.unscaleX(extent.x),
                                vp.unscaleY(extent.y + extent.h - 1)),
                         GPoint(vp.unscaleX(extent.x + extent.w - 1),
                                vp.unscaleY(extent.y)),
                         marked_only);
        for (int i = univ.ty_any_ships.findNextIndex(0); i != 0; i = univ.ty_any_ships.findNextIndex(i)) {
            const GShip& mate = univ.getShip(i);
            if (isValidChunnelMate(*chunnel_ship, mate)) {
                locker.addObject(mate);
            }
        }
        userMoveScannerTo(locker.getPoint());
    } else {
        WScannerChartWidget::doItemLock(clicked, mode, marked_only, optimize_warp);
    }
}

/** Double click. Same as F10. */
void
WShipNavigationChart::onDblClick()
{
    stop(cm_Enter);
}

static bool
handleConfirm(WShipNavigationChart& chart, GUniverse& univ, GShip& ship)
{
    if (chart.isChunnelMode()) {
        int fc = chooseChunnelMate(univ, ship, chart.getScannerTarget());
        if (fc == 0) {
            return false;
        }

        /* Clear speed and waypoint, set FC */
        setFleetWaypoint(univ, ship.getId(), ship.getPos());
        setFleetSpeed(univ, ship.getId(), 0);
        ship.setFCode(format("%03d", fc));

        /* Mate */
        GShip& mate = univ.getShip(fc);
        if (mate.isPlayable(GObject::Playable)) {
            /* For simplicity, use the fleet calls (although mates never are fleet members) */
            setFleetWaypoint(univ, fc, mate.getPos());
            setFleetSpeed(univ, fc, 0);
            // FIXME: if mate orbits a planet, PCC 1.x loads fuel onto it
        }

        /* Check whether chunnel works */
        GChunnelData cd;
        parseChunnelMission(ship, univ, cd);

        bool probFlag = false;
        string_t probStr = _("Please fix the following problems to make the chunnel work:");
        static const char*const problems[] = {
            N_("Initiator damaged"),
            N_("Initiator needs fuel"),
            N_("Initiator under tow"),
            N_("Mate damaged"),
            N_("Mate needs fuel"),
            N_("Mate moving"),
            N_("Mate under tow"),
        };
        static const int problemFlags[] = {
            GChunnelData::chf_Damaged,
            GChunnelData::chf_Fuel,
            GChunnelData::chf_Towed,
            GChunnelData::chf_MateDamaged,
            GChunnelData::chf_MateFuel,
            GChunnelData::chf_MateMoving,
            GChunnelData::chf_MateTowed,
        };
        for (uint_t i = 0; i < countof(problemFlags); ++i) {
            if (cd.failure & problemFlags[i]) {
                probStr += "\n" UTF_BULLET " ";
                probStr += _(problems[i]);
                probFlag = true;
            }
        }
        
        if (probFlag) {
            messageBox(probStr, _("Chunnel"));
        }
        return true;
    } else {
        setFleetWaypoint(univ, ship.getId(), chart.getScannerTarget());
        /* FIXME: setOptimumWarp does not belong here. Better do it in the script. But how? */
        setOptimumWarp(univ, ship.getId());
        if (ship.isFleetLeader()) {
            synchronizeFleet(univ, ship.getId());
        }
        return true;
    }
}

/** Display ship navigation chart. This will set the ship's waypoint
    and/or fcode.
    \param univ Universe we're in
    \param ship Ship to work on */
void
doShipNavigationChart(GUniverse& univ, GShip& ship)
{
    int size = ui_root->getExtent().h;

    // VBox
    //   UIFrameGroup
    //     WShipNavigationChart
    //   HBox
    //     WScanResult
    //     VBox
    //       UISpacer  -or-  HBox / Chunnel "C"
    //       HBox
    //         "OK"
    //         "ESC"
    //         "Help"

    UIWindow win(_("Navigation Chart"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);

    ObjectHolder h;

    UIGroup& g2   = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g22  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g222 = h.add(new UIGroup(UIHBoxLayout::instance5));

    WShipNavigationChart& chart = h.add(new WShipNavigationChart(univ));
    chart.setCenterPosition(ship.getPos());
    chart.setScannerOrigin(ship.getPos());
    chart.setScannerTarget(ship.getWaypoint());
    chart.setState(chart.sc_StyleChange | chart.sc_Zoomable | chart.sc_Keyboard, true);
    chart.setCircleRadius(ship.isHyperdriving(),
                          square(ship.getWarp()) * (ship.canDoSpecial(hf_Gravitonic) ? 2 : 1));
    chart.sig_dblclick.add(&chart, &WShipNavigationChart::onDblClick);
    win.add(UIFrameGroup::wrap(h, chart));

    WScanResult& scanner = h.add(new WScanResult(false));
    scanner.setChart(chart);
    g2.add(scanner);
    g2.add(g22);

    /* Offer chunneling when ship is not member of a fleet only */
    if (ship.getFleetNumber() == 0
        && (ship.canDoSpecial(hf_FirecloudChunnel)
            || ship.canDoSpecial(hf_ChunnelSelf)
            || ship.canDoSpecial(hf_ChunnelOthers)))
    {
        UIGroup& g221 = h.add(new UIGroup(UIHBoxLayout::instance5));
        UIButton& btn = h.add(new UIButton(0, "C", 'c'));
        btn.setState(UIButton::bf_SmallSize, true);
        g221.add(h.add(new UISpacer()));
        g221.add(h.add(new UIStaticText(_("Make Chunnel"), UIStaticText::StaticNormal)));
        g221.add(btn);
        g22.add(g221);
        chart.setChunnelButton(&btn);
        chart.setChunnelShip(ship);
        btn.addCall(&chart, &WShipNavigationChart::toggleChunnel);
    } else {
        g22.add(h.add(new UISpacer()));
    }
    g22.add(g222);

    g222.add(h.add(new UIButton(cm_Enter, _("F10 - OK"), SDLK_F10)).addStop());
    g222.add(h.add(new UIButton(cm_Escape, "ESC", SDLK_ESCAPE)).addStop());
    g222.add(h.add(new UIButton("H", 'h')));
    
    win.add(g2);
    win.add(h.add(new WHelpWidget("pcc2:navchart")));
    win.setExtent(GfxRect(0, 0, size - 50, size - 10));

    ui_root->centerWidget(win);
    ui_root->add(win);
    while (ui_root->run() == cm_Enter) {
        if (handleConfirm(chart, univ, ship)) {
            break;
        }
    }
    ui_root->close(win);
}


/** Display generic navigation chart. This will allow the user to
    select a point, but is not bound to a specific ship. Therefore, it
    will not allow setting up chunnels.
    \param univ [in] Universe
    \param origin [in] Origin of selection (location of ship)
    \param target [in/out] Initially-selected waypoint
    \param hyp [in] true to show hyp-circles, false for regular warp circles
    \param r [in] regular warp circle radius, if any */
bool
doGenericNavigationChart(GUniverse& univ, GPoint origin, GPoint& target, bool hyp, int r)
{
    int size = ui_root->getExtent().h;

    // VBox
    //   UIFrameGroup
    //     WShipNavigationChart
    //   HBox
    //     WScanResult
    //     VBox
    //       UISpacer
    //       HBox
    //         "OK"
    //         "ESC"
    //         "Help"

    UIWindow win(_("Select Waypoint"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
    win.setLayout(UIVBoxLayout::instance5);

    ObjectHolder h;

    UIGroup& g2   = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g22  = h.add(new UIGroup(UIVBoxLayout::instance5));
    UIGroup& g222 = h.add(new UIGroup(UIHBoxLayout::instance5));

    WShipNavigationChart& chart = h.add(new WShipNavigationChart(univ));
    chart.setCenterPosition(origin);
    chart.setScannerOrigin(origin);
    chart.setScannerTarget(target);
    chart.setState(chart.sc_StyleChange | chart.sc_Zoomable | chart.sc_Keyboard, true);
    chart.setCircleRadius(hyp, r);
    chart.sig_dblclick.add(&chart, &WShipNavigationChart::onDblClick);
    win.add(UIFrameGroup::wrap(h, chart));

    WScanResult& scanner = h.add(new WScanResult(false));
    scanner.setChart(chart);
    g2.add(scanner);
    g2.add(g22);
    g22.add(h.add(new UISpacer()));
    g22.add(g222);

    g222.add(h.add(new UIButton(cm_Enter, _("F10 - OK"), SDLK_F10)).addStop());
    g222.add(h.add(new UIButton(cm_Escape, "ESC", SDLK_ESCAPE)).addStop());
    g222.add(h.add(new UIButton("H", 'h')));
    
    win.add(g2);
    win.add(h.add(new WHelpWidget("pcc2:navchart")));
    win.setExtent(GfxRect(0, 0, size - 50, size - 10));

    ui_root->centerWidget(win);
    if (ui_root->runModal(win) == cm_Enter) {
        target = chart.getScannerTarget();
        return true;
    } else {
        return false;
    }
}
