/**
  *  \file client/dialogs/simoptions.cc
  *  \brief Simulator Options Dialog
  */

#include "client/dialogs/simoptions.h"
#include "client/widgets/helpwidget.h"
#include "client/wtile.h"
#include "cpluslib/format.h"
#include "game/sim.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/uiglobal.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/decimalsel.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/strlist.h"
#include "util/strtab.h"

namespace {
    class WSimOptionTile : public WTile {
        GSimOptions& opts;
        ObjectHolder h;

        static const char*const left_column[];
        static const char*const right_column[];
        static const char keys[];

        int left_x, left_wi;
        int right_x;

        void chooseMode();
        void computeLayout();
     public:
        WSimOptionTile(GSimOptions& opts);

        bool handleEvent(const UIEvent& e, bool second_pass);
        void getLayoutInfo(LayoutInfo& info);
        void drawData(GfxCanvas& can);
    };
}

/* Strings are enumerated here to simplify layout computation */
const char*const WSimOptionTile::left_column[] = {
    N_("Engine-Shield Bonus:"),
    N_("Fed Crew Bonus:"),
    N_("Random Left/Right:"),
    N_("Left/Right Balance:"),
    N_("VCR Algorithm:"),
    N_("Honor alliances/teams:"),
    N_("Battles per simulation:"),
    N_("Seed control:"),
    N_("Randomize FCodes:")
};

const char WSimOptionTile::keys[] = { 'E', 'C', 'R', 'L', 'P', 'A', 'B', 'S', 'F' };

/* Note: for simplicity, we assume that the first entries
   match VcrMode */
const char*const WSimOptionTile::right_column[] = {
    /* Host mode */
    N_("Host"),                 // 0
    N_("PHost 2"),
    N_("PHost 3"),
    N_("PHost 4"),
    N_("FLAK"),
    N_("NuHost"),               // 5
    0,
    0,
    0,
    0,

    /* ES-Bonus */
    N_("none"),                 // 10
    /* Boolean options */
    N_("Yes"),                  // 11
    N_("No"),                   // 12
    /* Balancing */
    N_("360 kt (Host)"),        // 13
    N_("Master at Arms"),       // 14
    /* battles */
    N_("one"),                  // 15
    N_("complete"),             // 16
    /* Random fcode */
    N_("once"),                 // 17
    N_("every time")            // 18
};

WSimOptionTile::WSimOptionTile(GSimOptions& opts)
    : WTile(GfxPoint(1, 1)), opts(opts)
{
    using std::tolower;
    const int ly = font_heights[FONT_TITLE];
    setLayout(LeftJustified, 0, false);
    for (uint_t i = 0; i < countof(keys); ++i) {
        addControl(GfxRect(0, ly*i+1, ly-2, ly-2),
                   h.add(new UIButton(string_t(&keys[i], 1),
                                      tolower(keys[i]))));
    }
}

bool
WSimOptionTile::handleEvent(const UIEvent& e, bool second_pass)
{
    if (e.type == ev_Key) {
        switch (e.key.code) {
         case 'e':
            // ESB
            {
                UIDecimalSelector select(0, 0, 500, 10);
                select.setValue(opts.engine_shield_bonus);
                if (select.doStandardDialog(_("Engine-Shield Bonus"), _("Rate [%]"), 0)) {
                    opts.engine_shield_bonus = select.getValue();
                    drawWidget();
                }
            }
            return true;
         case 'c':
            // Fed bonus
            opts.scotty_bonus = !opts.scotty_bonus;
            drawWidget();
            return true;
         case 'r':
            // random l/r
            opts.random_left_right = !opts.random_left_right;
            drawWidget();
            return true;
         case 'l':
            // balancing
            if (int(opts.balancing_mode) >= opts.BalanceMAX) {
                opts.balancing_mode = GSimOptions::BalancingMode(0);
            } else {
                opts.balancing_mode = GSimOptions::BalancingMode(opts.balancing_mode + 1);
            }
            drawWidget();
            return true;
         case 'p':
            // host
            chooseMode();
            drawWidget();
            return true;
         case 'a':
            // allies
            opts.honor_alliances = !opts.honor_alliances;
            drawWidget();
            return true;
         case 'b':
            // battles per sim
            opts.only_one_simulation = !opts.only_one_simulation;
            if (!opts.only_one_simulation)
                opts.seed_control = false;
            drawWidget();
            return true;
         case 's':
            // seed control
            opts.seed_control = !opts.seed_control;
            if (opts.seed_control) {
                opts.only_one_simulation = true;
            }
            drawWidget();
            return true;
         case 'f':
            // random fcodes
            opts.randomize_fcodes_on_every_fight = !opts.randomize_fcodes_on_every_fight;
            drawWidget();
            return true;
        }
    }
    return WTile::handleEvent(e, second_pass);
}

void
WSimOptionTile::getLayoutInfo(LayoutInfo& info)
{
    computeLayout();
    info.min_size = info.pref_size =
        GfxPoint(right_x + 5, countof(left_column) * font_heights[FONT_TITLE]);
    info.flags = 0;
}

void
WSimOptionTile::drawData(GfxCanvas& can)
{
    // draw left side
    computeLayout();

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(getSkin());
    ctx.useFont(*fonts[FONT_NORMAL]);
    ctx.setTextAlign(0, 1);
    ctx.setColor(UIColor::tc_Static);
    for (unsigned int i = 0; i < countof(left_column); ++i) {
        int y = (font_heights[FONT_TITLE] * (2*i+1)) / 2;
        outText(ctx, left_x, y, _(left_column[i]));
    }

    // draw right side
    int right_wi = right_x - left_x - left_wi;
    ctx.setTextAlign(2, 1);

    const int ly = font_heights[FONT_TITLE];
    int y = ly/2;

    ctx.setColor(UIColor::tc_Green);
    ctx.setSolidBackground();

    // ESB
    if (opts.engine_shield_bonus != 0)
        outTextF(ctx, right_x, y, right_wi, format("%d%%", opts.engine_shield_bonus));
    else
        outTextF(ctx, right_x, y, right_wi, _(right_column[10]));
    y += ly;

    // Crew Bonus
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.scotty_bonus ? 11 : 12]));
    y += ly;

    // Random left/right
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.random_left_right ? 11 : 12]));
    y += ly;

    // Left/right balance
    if (opts.balancing_mode == GSimOptions::Balance360k)
        outTextF(ctx, right_x, y, right_wi, _(right_column[13]));
    else if (opts.balancing_mode == GSimOptions::BalanceMasterAtArms)
        outTextF(ctx, right_x, y, right_wi, _(right_column[14]));
    else
        outTextF(ctx, right_x, y, right_wi, _(right_column[10]));
    y += ly;

    // VCR Algorithm
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.vcr_mode]));
    y += ly;

    // Honor alliances
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.honor_alliances ? 11 : 12]));
    y += ly;

    // Battles per simulation
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.only_one_simulation ? 15 : 16]));
    y += ly;

    // Seed control
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.seed_control ? 11 : 12]));
    y += ly;

    // Random fcodes
    outTextF(ctx, right_x, y, right_wi, _(right_column[opts.randomize_fcodes_on_every_fight ? 18 : 17]));
    y += ly;
}

void
WSimOptionTile::computeLayout()
{
    int right_wi;

    left_wi = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(left_column, countof(left_column))) + 10;
    right_wi = getTextWidth(*fonts[FONT_NORMAL], StaticTranslatedStringTable(right_column, countof(right_column)));

    left_x = font_heights[FONT_TITLE] + 5;
    right_x = left_x + left_wi + right_wi + 5;
}

void
WSimOptionTile::chooseMode()
{
    // Build list of modes
    StringList sl;
    for (int i = 0; i <= GSimOptions::VcrMAX; ++i) {
        // do we accept this mode?
#ifndef CONF_FLAK_SUPPORT
        if (i == GSimOptions::VcrFLAK)
            continue;
#endif
        sl.add(i, _(WSimOptionTile::right_column[i]));
    }

    // Ask user
    UIStandardListbox box(0, sl.getCount(), getTextWidth(*fonts[FONT_NORMAL], sl) + 20, sl);
    box.scrollToKey(opts.vcr_mode);
    if (box.doStandardListbox(_("VCR Algorithm"))) {
        opts.vcr_mode = GSimOptions::VcrMode(box.getKey());
    }
}

/** Edit simulator options.
    \param opts [in/out] Options
    \retval true Options were possibly changed, dialog confirmed
    \retval false Options not changed, dialog cancelled */
bool
editSimulatorOptions(GSimOptions& opts)
{
    // Temporary copy to work on
    GSimOptions work_opts = opts;

    // Build dialog
    ObjectHolder h;

    UIWindow& win(h.add(new UIWindow(_("Simulator Options"), UIWindow::BLUE_WINDOW, COLOR_GRAY)));
    win.setLayout(UIVBoxLayout::instance5);
    win.add(h.add(new WSimOptionTile(work_opts)));

    UIGroup& group_buttons(h.add(new UIGroup(UIHBoxLayout::instance5)));
    group_buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
    group_buttons.add(h.add(new UIButton(cm_Enter, _("Cancel"), SDLK_ESCAPE)).addStop());
    group_buttons.add(h.add(new UISpacer()));
    group_buttons.add(h.add(new UIButton(_("Help"), 'h')));
    win.add(group_buttons);
    win.add(h.add(new WHelpWidget("pcc2:simopts")));
    win.pack();
    ui_root->centerWidget(win);

    // Run it
    int rv = ui_root->runModal(win);
    if (rv == cm_Enter) {
        opts = work_opts;
        return true;
    } else {
        return false;
    }
}
