/**
  *  \file client/dialogs/simplanettorps.cc
  *  \brief Simulator Planetary Torpedoes Editor
  */

#include "client/dialogs/simplanettorps.h"
#include "game/sim.h"
#include "game/spec.h"
#include "gfx/gfxcontext.h"
#include "gfx/palette.h"
#include "ui/commands.h"
#include "ui/layout.h"
#include "ui/skin.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/widgets/decimalsel.h"
#include "ui/widgets/focusiterator.h"
#include "ui/window.h"
#include "util/misc.h"
#include "util/nls.h"

namespace {
    class WSimPlanetTorpLine : public UIDecimalSelector {
        const int torpnr;
     public:
        WSimPlanetTorpLine(int torpnr, int value);
        ~WSimPlanetTorpLine();
        void drawContent(GfxCanvas& can);
        void getLayoutInfo(LayoutInfo& info);
    };
}

WSimPlanetTorpLine::WSimPlanetTorpLine(int torpnr, int value)
    : UIDecimalSelector(torpnr, 0, 10000, 10),
      torpnr(torpnr)
{
    setValue(value);
}

WSimPlanetTorpLine::~WSimPlanetTorpLine()
{ }

void
WSimPlanetTorpLine::drawContent(GfxCanvas& can)
{
    string_t text_l = " " + getTorp(torpnr).getName();
    string_t text_r = numToString(getValue()) + " ";
    const UIColorSet& scheme = (getFocusState() != NoFocus ? BLACK_COLOR_SET : GRAY_COLOR_SET);

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);
    ctx.useFont(*fonts[FONT_NORMAL]);

    // Background
    drawSolidBar(ctx, extent, scheme[UIColor::tc_Background]);

    // length
    int len   = fonts[FONT_NORMAL]->getTextWidth(text_r);
    int split = extent.w - len;

    // draw right side first
    ctx.setColor(scheme[UIColor::tc_Static]);
    outTextF(ctx, extent.x + split, extent.y, len, text_r);
    outTextF(ctx, extent.x, extent.y, split, text_l);

    can.update(extent);
}

void
WSimPlanetTorpLine::getLayoutInfo(LayoutInfo& info)
{
    info.min_size = info.pref_size =
        GfxPoint(fonts[FONT_NORMAL]->getTextWidth(getTorp(torpnr).getName() + "  99999"),
                 font_heights[FONT_NORMAL]);
    info.flags = li_FlexibleH;
}

/** Edit torpedoes in a simulator planet.
    \param planet The planet
    \retval true Dialog was confirmed, planet possibly changed
    \retval false Dialog was cancelled, planet not changed */
bool
editSimulatorPlanetTorps(GSimPlanet& planet)
{
    WSimPlanetTorpLine* torp_edits[NUM_TORPS];
    ObjectHolder h;
    UIWindow& w = h.add(new UIWindow(_("Starbase Torpedoes"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    w.setLayout(UIVBoxLayout::instance5);

    UIGroup& group_torps   = h.add(new UIGroup(UIVBoxLayout::instance0));
    UIGroup& group_buttons = h.add(new UIGroup(UIHBoxLayout::instance5));
    for (int i = 1; i <= NUM_TORPS; ++i) {
        WSimPlanetTorpLine& tl = h.add(new WSimPlanetTorpLine(i, planet.getBaseTorps(i)));
        torp_edits[i-1] = &tl;
        group_torps.add(tl);
    }
    group_torps.add(h.add(new UIFocusIterator(1, NUM_TORPS,
                                              UIFocusIterator::fi_Vert | UIFocusIterator::fi_Page | UIFocusIterator::fi_Home)));
    group_buttons.add(h.add(new UIButton(cm_Enter, _("OK"), SDLK_RETURN)).addStop());
    group_buttons.add(h.add(new UIButton(cm_Escape, _("Cancel"), SDLK_ESCAPE)).addStop());
    group_buttons.add(h.add(new UISpacer()));
    w.add(group_torps);
    w.add(group_buttons);
    int focus_id = planet.getBaseTorpTech();
    torp_edits[focus_id-1]->setFocusedWidget();

    w.pack();
    ui_root->moveWidgetToEdge(w, 2, 2, 10);
    int rv = ui_root->runModal(w);
    if (rv == cm_Enter) {
        for (int i = 1; i <= NUM_TORPS; ++i) {
            planet.setBaseTorps(i, torp_edits[i-1]->getValue());
        }
        return true;
    } else {
        return false;
    }
}
