/**
  *  \file client/dialogs/weaponeff.cc
  *  \brief Weapon Effects Dialog
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/weaponeff.h"
#include "game/shipquery.h"
#include "cpluslib/holder.h"
#include "ui/window.h"
#include "ui/layout.h"
#include "ui/widgets/button.h"
#include "ui/richdocument.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"
#include "cpluslib/format.h"
#include "util/nls.h"
#include "game/racename.h"
#include "ui/widgets/keytranslator.h"
#include "game/spec.h"
#include "game/host.h"
#include "util/misc.h"
#include "util/unicodechars.h"
#include "game/formula.h"
#include "game/config.h"

/** Show integer effect.
    \param d Document
    \param x Center here
    \param eff Effect (=damage)
    \param limit Limit (=max total damage) */
static void
showEffectI(RichDocument& d, int x, int eff, int limit)
{
    if (eff == 0) {
        d.addCentered(x, RichText(UIColor::tc_Faded, "-"));
    } else {
        d.addRight(x, numToString(eff));
        d.add(format(" (%d" UTF_TIMES ")", (limit + eff-1) / eff));
    }
}

/** Show float effect.
    \param d Document
    \param x Center here
    \param eff Effect (=damage)
    \param limit Limit (=max total damage) */
static void
showEffectF(RichDocument& d, int x, double eff, int limit)
{
    if (eff == 0) {
        d.addCentered(x, RichText(UIColor::tc_Faded, "-"));
    } else {
        d.addRight(x, string_t(format("%.2f", eff)));
        d.add(format(" (%d" UTF_TIMES ")", int((limit / eff) + 0.999)));
    }
}

/** Show effect for one weapon.
    \param d Document
    \param query Ship to query
    \param kill Kill power of weapon
    \param expl Explosive power of weapon
    \param deathRay true to handle death rays */
static void
showWeaponEffect(RichDocument& d, const GShipQuery& query, int kill, int expl, bool deathRay)
{
    const int em = font_ems[FONT_NORMAL];
    const int damageLimit = raceId(query.getOwner()) == 2 ? 151 : 100;
    if (!host.isPHost()) {
        /* Tim-Host */
        int mass = query.getCombatMass() + 1;
        int shieldEff = radiv(80*expl,      mass) + 1;
        int hullEff   = radiv(80*shieldEff, mass) + 1;
        int crewEff   = radiv(80*kill,      mass);
        showEffectI(d, 17*em, shieldEff, 100);
        showEffectI(d, 25*em, hullEff, damageLimit);
        showEffectI(d, 33*em, crewEff, query.getCrew());
    } else {
        bool isDeathRay = deathRay && host.hasDeathRays() && expl == 0;
        if (config.AllowAlternativeCombat()) {
            double mass = query.getCombatMass() + 1.0;
            double shieldEff = (config.ShieldDamageScaling(1) * int32_t(expl)
                                + config.ShieldKillScaling(1) * int32_t(kill)) / mass;
            double hullEff = int32_t(expl) * config.HullDamageScaling(1) / mass;
            double crewEff = int32_t(kill) * config.CrewKillScaling(1) / mass;
            if (isDeathRay)
                shieldEff = hullEff = 0.0;
            showEffectF(d, 17*em, shieldEff, 100);
            showEffectF(d, 25*em, hullEff, damageLimit);
            showEffectF(d, 33*em, crewEff, query.getCrew());
        } else {
            int mass = query.getCombatMass() + 1;
            int shieldEff = radiv(config.ShieldDamageScaling(1) * int32_t(expl)
                                  + config.ShieldKillScaling(1) * int32_t(kill),
                                  mass) + 1;
            int hullEff = radiv(shieldEff * config.ShieldDamageScaling(1), mass);
            int crewEff = radiv(kill * config.CrewKillScaling(1), mass);
            if (isDeathRay) {
                shieldEff = hullEff = 0;
                if (crewEff == 0)
                    crewEff = 1;
            }
            showEffectI(d, 17*em, shieldEff, 100);
            showEffectI(d, 25*em, hullEff, damageLimit);
            showEffectI(d, 33*em, crewEff, query.getCrew());
        }
    }
}

/** Show weapon effects dialog for one ship query. */
void
showWeaponEffects(const GShipQuery& query)
{
    /* We can only display this dialog when we have ship mass, crew, and owner */
    if (query.getCombatMass() <= 0 || query.getOwner() <= 0 || query.getCrew() <= 0)
        return;

    /* Build dialog widgets */
    // 3 Lines of header,
    // 10 beams
    // 10 torps
    // fighters
    // gaps
    // => 25 lines

    ObjectHolder h;
    UIWindow& win = h.add(new UIWindow(_("Weapon Effects"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
    win.setLayout(UIVBoxLayout::instance5);

    UIRichDocument& doc = h.add(new UIRichDocument(font_sizes[FONT_NORMAL].scaledBy(40, 25), 0));
    win.add(doc);

    UIGroup& g = h.add(new UIGroup(UIHBoxLayout::instance5));
    g.add(h.add(new UISpacer()));
    g.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g.add(h.add(new UISpacer()));
    win.add(g);
    win.add(h.add(new UIKeyTranslator()).addStop(' ', 1).addStop(SDLK_ESCAPE, 1));
    win.pack();
    ui_root->centerWidget(win);

    /* Prepare text */
    const int em = font_ems[FONT_NORMAL];
    RichDocument& d = doc.getDocument();
    d.add(format(_("Effects on %d kt %s ship"), query.getCombatMass(), player_racenames.getAdjName(query.getOwner())));
    if (query.getUsedESB())
        d.add(format(_(" (using %d%% ESB)"), query.getUsedESB()));
    d.add(format(_(", %d crewm%1{a%|e%}n"), query.getCrew()));
    d.addParagraph();
    d.addCentered(17*em, RichText(_("Shield")).withAttribute(RichTextStyleAttribute::Bold));;
    d.addCentered(25*em, RichText(_("Hull")).withAttribute(RichTextStyleAttribute::Bold));;
    d.addCentered(33*em, RichText(_("Crew")).withAttribute(RichTextStyleAttribute::Bold));;
    d.addNewline();
    for (int i = 1; i <= NUM_BEAMS; ++i) {
        d.add(getBeam(i).getName());
        showWeaponEffect(d, query, getBeam(i).getKill(), getBeam(i).getDamage(), true);
        d.addNewline();
    }
    d.addNewline();
    int scale = (host.isPHost() && config.AllowAlternativeCombat()) ? 1 : 2;
    for (int i = 1; i <= NUM_TORPS; ++i) {
        d.add(getTorp(i).getName());
        showWeaponEffect(d, query, scale*getTorp(i).getKill(), scale*getTorp(i).getDamage(), true);
        d.addNewline();
    }
    d.addNewline();
    d.add(_("Fighter Beam"));
    if (host.isPHost()) {
        showWeaponEffect(d, query, config.FighterBeamKill(1), config.FighterBeamExplosive(1), false);
    } else {
        showWeaponEffect(d, query, 2, 2, false);
    }
    d.finish();

    /* Do it */
    ui_root->runModal(win);
}
