/**
  *  \file flak/vis.cc
  *  \brief FLAK: Visualisation
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "flak/vis.h"
#include "cpluslib/assert.h"
#include "gfx/palette.h"
#include "game/config.h"
#include "flak/vis-sprites.h"
#include "util/misc.h"

namespace {
    /** Status box display. This sprite displays a box around a ship.
        It automatically follows its ship. */
    class FlakBoxSprite : public GfxSprite3D {
        enum {
            corner_size = 2,
            bar_height  = 4,
            bar_dist    = bar_height + 1
        };

        FlakFancyVisualizer& vis;
        FlakBattle::ship_t ship;
     public:
        FlakBoxSprite(FlakFancyVisualizer& vis, FlakBattle::ship_t ship);
        void tick();
        void draw(GfxCanvas& can);
    };
}

FlakBoxSprite::FlakBoxSprite(FlakFancyVisualizer& vis, FlakBattle::ship_t ship)
    : GfxSprite3D(FlakFancyVisualizer::layer_Boxes), vis(vis), ship(ship)
{
    tick();
}

void
FlakBoxSprite::tick()
{
    Ptr<GfxSprite3D> sp = vis.getShipSprite(ship);
    if (!sp || !vis.getBattle().getShipByNumber(ship).isAlive()) {
        vis.setShipStatus(ship, false);
    } else {
        // FIXME: check whether we actually need to update the sprite to
        // save time in case the ship does not move and not change
        sp->ensureTick();
        GfxRect a = sp->getPos();
        a.x -= 2;
        a.w += 4;
        a.y -= bar_dist;
        a.h += bar_dist + 2;
        setPos(a, sp->getPosZ());

        update();
    }
}

void
FlakBoxSprite::draw(GfxCanvas& can)
{
    const FlakShip& sh = vis.getBattle().getShipByNumber(ship);

    // corners
    GfxRect r = getPos();

    const uint32 w = standard_colors[COLOR_WHITE];
    const uint32 b = standard_colors[COLOR_BLACK];

    // top-left corner
    can.drawHLine(GfxPoint(r.x, r.y + bar_dist),      corner_size,   w, 255U, GFX_ALPHA_OPAQUE);
    can.drawVLine(GfxPoint(r.x, r.y + bar_dist + 1),  corner_size-1, w, 255U, GFX_ALPHA_OPAQUE);
    // top-right
    can.drawHLine(GfxPoint(r.x + r.w - corner_size, r.y + bar_dist),     corner_size,   w, 255U, GFX_ALPHA_OPAQUE);
    can.drawVLine(GfxPoint(r.x + r.w - 1,           r.y + bar_dist + 1), corner_size-1, w, 255U, GFX_ALPHA_OPAQUE);
    // bottom-left
    can.drawHLine(GfxPoint(r.x, r.y + r.h - 1),           corner_size,   w, 255U, GFX_ALPHA_OPAQUE);
    can.drawVLine(GfxPoint(r.x, r.y + r.h - corner_size), corner_size-1, w, 255U, GFX_ALPHA_OPAQUE);
    // bottom-right
    can.drawHLine(GfxPoint(r.x + r.w - corner_size, r.y + r.h - 1),           corner_size,   w, 255U, GFX_ALPHA_OPAQUE);
    can.drawVLine(GfxPoint(r.x + r.w - 1,           r.y + r.h - corner_size), corner_size-1, w, 255U, GFX_ALPHA_OPAQUE);

    // shield bar
    int total_pixels = 2 * r.w + 2 * bar_height - 4; // '4' means four corners
    int shield_pixels = int(total_pixels * sh.status.shield / 100);

    GfxRect foo = GfxRect(r.x, r.y, r.w, bar_height);

    // left; white pixels are on top
    int total = foo.h - 1;
    int now   = std::min(shield_pixels, total);
    can.drawVLine(GfxPoint(foo.x, foo.y),       now,       w, 255U, GFX_ALPHA_OPAQUE);
    can.drawVLine(GfxPoint(foo.x, foo.y + now), total-now, b, 255U, GFX_ALPHA_OPAQUE);
    shield_pixels -= now;

    // bottom; white pixels are on the left
    total = foo.w - 1;
    now   = std::min(shield_pixels, total);
    can.drawHLine(GfxPoint(foo.x,       foo.y + foo.h - 1), now,       w, 255U, GFX_ALPHA_OPAQUE);
    can.drawHLine(GfxPoint(foo.x + now, foo.y + foo.h - 1), total-now, b, 255U, GFX_ALPHA_OPAQUE);
    shield_pixels -= now;

    // right; white pixels are on bottom
    total = foo.h - 1;
    now   = std::min(shield_pixels, total);
    can.drawVLine(GfxPoint(foo.x + foo.w - 1, foo.y + foo.h - now), now,       w, 255U, GFX_ALPHA_OPAQUE);
    can.drawVLine(GfxPoint(foo.x + foo.w - 1, foo.y + 1),           total-now, b, 255U, GFX_ALPHA_OPAQUE);
    shield_pixels -= now;

    // top; white pixels are right
    total = foo.w - 1;
    now   = std::min(shield_pixels, total);
    can.drawHLine(GfxPoint(foo.x + foo.w - now, foo.y), now,       w, 255U, GFX_ALPHA_OPAQUE);
    can.drawHLine(GfxPoint(foo.x + 1,           foo.y), total-now, b, 255U, GFX_ALPHA_OPAQUE);

    // damage bar
    foo.grow(-1, -1);
    int limit = raceId(sh.data.player)==2 ? 150 : 100;
    int divi = int(foo.w - foo.w * sh.status.damage / limit);
    int color = int(15 - 15 * sh.status.damage / limit);
    if (divi < 0)
        divi = 0;
    if (color < 0)
        color = 0;
    can.drawBar(GfxRect(foo.x,        foo.y, divi,         foo.h), standard_colors[COLOR_STATUS+color], 0, SOLID_FILL, GFX_ALPHA_OPAQUE);
    can.drawBar(GfxRect(foo.x + divi, foo.y, foo.w - divi, foo.h), standard_colors[COLOR_BLACK],        0, SOLID_FILL, GFX_ALPHA_OPAQUE);
}

/************************** FlakFancyVisualizer **************************/

const int32 FlakFancyVisualizer::EndOfFight;

/** Create new visualizer.
    \param battle battle to display
    \param area   area for battlefield */
FlakFancyVisualizer::FlakFancyVisualizer(FlakBattle& battle, GfxRect area)
    : FlakVisualizer(battle), manager(area),
      num_fx_sprites(0), num_explosion_sprites(0),
      any_ship_changed(false), is_time_advance(false),
      ship_info(battle.getNumShips()),
      status_time_div(20)
{ }

FlakFancyVisualizer::~FlakFancyVisualizer()
{
    safeClear(status_tokens);
}

/** Set object sprite.
    \param o         object
    \param sprite    associated sprite. Can be null to remove the association.
    \param is_new    true iff this is a new object, false if you already called
                     setObjectSprite() on it. */
void
FlakFancyVisualizer::setObjectSprite(FlakObject& o, Ptr<GfxSprite3D> sprite, bool is_new)
{
    if (o.vis_data != id_assoc.nil && !is_new) {
        id_assoc.getObject(o.vis_data)->deleteThis();
        id_assoc.removeObject(o.vis_data);
    }

    if (sprite) {
        o.vis_data = id_assoc.addObject(sprite);
        manager.addSprite(sprite);
    } else {
        o.vis_data = id_assoc.nil;
    }
}

/** Mark ship changed. Use this when you detect that a ship changed. */
void
FlakFancyVisualizer::markShipChanged(ship_t ship)
{
    ship_info[ship].changed = true;
    any_ship_changed = true;
}

/** Swap status token log with other visualizer. */
void
FlakFancyVisualizer::swapStatusTokens(FlakFancyVisualizer& other)
{
    std::swap(status_time_div, other.status_time_div);
    status_tokens.swap(other.status_tokens);
}

/** Initialize visualizer.
    \param is_startup=true this is the initialisation of the visualisation
    \param is_startup=false position within fight changed */
void
FlakFancyVisualizer::init(bool is_startup)
{
    // initialize battle if needed
    if (is_startup)
        getBattle().init();

    // clear objects
    id_assoc.clear();

    // keep scene and text
    manager.deleteSpritesWithId(layer_Torp);
    manager.deleteSpritesWithId(layer_Fighter);
    manager.deleteSpritesWithId(layer_Beam);
    manager.deleteSpritesWithId(layer_Bang);
    manager.deleteSpritesWithId(layer_Ships);

    // create everything anew
    createSprites(is_startup);
}

/** Change time. This advances / rewinds the time in the FlakBattle,
    and performs necessary sprite operations. Caller must still call
    doUpdates(), though.
    \param new_time new time. If too small or too large, we move to
    the closest possible time. */
void
FlakFancyVisualizer::setTime(int32 new_time)
{
    if (new_time < 0)
        new_time = 0;

    if (new_time == the_battle.getTime()) {
        // ok
    } else if (new_time == the_battle.getTime() + 1) {
        // play one normal tick
        /* The ultra-clean way would have been to tick() until all fx
           sprites are gone. This does not work, though, as deletions
           physically happen at doUpdates() which we cannot call since
           we don't have a canvas.,  */
        getManager().deleteSpritesWithId(layer_Beam);
        is_time_advance = true;
        if (!the_battle.isTerminated()) {
            the_battle.playOneTick();
            manager.tick();
            maybeSaveStatus();
        }
        is_time_advance = false;
    } else {
        // figure closest checkpoint
        enum How { hInit, hFastForward, hCheckpoint };

        // assume closest checkpoint is rewind
        int32 cptime = 0;
        How   how = hInit;

        // faster by fast-forward?
        if (new_time > the_battle.getTime()) {
            cptime = the_battle.getTime();
            how = hFastForward;
        }

        // checkpoints
        std::size_t slot = new_time / status_time_div;
        if (slot != 0 && status_tokens.size() > 1) {
            if (slot >= status_tokens.size())
                slot = status_tokens.size() - 1;
            ASSERT(status_tokens[slot] != 0);
            ASSERT(slot * status_time_div <= new_time);
            if (slot * status_time_div > cptime) {
                cptime = slot * status_time_div;
                how = hCheckpoint;
            }
        }

        // go to found checkpoint
        switch (how) {
         case hInit:
            getBattle().init();
            init(false);
            manager.tick();
            break;
         case hFastForward:
            break;
         case hCheckpoint:
            status_tokens[slot]->storeTo(the_battle);
            init(false);
            break;
        }

        // play
        is_time_advance = true;
        while (the_battle.getTime() < new_time && !the_battle.isTerminated()) {
            the_battle.playOneTick();
            manager.tick();
            maybeSaveStatus();
        }
        is_time_advance = false;
    }
}

void
FlakFancyVisualizer::maybeSaveStatus()
{
    // save status if necessary
 try_again:
    if (the_battle.getTime() % status_time_div == 0) {
        int slot = the_battle.getTime() / status_time_div;
        if (slot >= status_max_size) {
            // increase divider
            std::size_t out = 0;
            for (std::size_t in = 0; in < status_tokens.size(); ++in) {
                // FIXME: language lawyer alert
                if (in % 2U == 0)
                    status_tokens[out++] = status_tokens[in];
                else
                    delete status_tokens[in];
            }
            status_tokens.erase(status_tokens.begin() + out, status_tokens.end());
            status_time_div *= 2;
            goto try_again;
        }
        while (status_tokens.size() <= slot)
            status_tokens.push_back(0);

        if (status_tokens[slot] == 0) {
            status_tokens[slot] = new FlakStatusToken(the_battle);
        }
    }
}

/** Do one playback tick. */
void
FlakFancyVisualizer::doTick(GfxCanvas& can)
{
    // clear all changed flags
    for (std::vector<ShipInfo>::iterator i = ship_info.begin(); i != ship_info.end(); ++i)
        i->changed = false;

    // do appropriate animation tick
    is_time_advance = true;
    if (num_fx_sprites) {
        manager.tick();
        manager.doUpdates(can);
    } else if (!the_battle.isTerminated()) {
        everyTickAction();
        the_battle.playOneTick();
        manager.tick();
        manager.doUpdates(can);
    } else if (num_explosion_sprites) {
        manager.tick();
        manager.doUpdates(can);
    }
    is_time_advance = false;

    maybeSaveStatus();
}

/** Change status of a ship.
    \param ship    ship number (0-based)
    \param boxed   true iff this ship should have a status box */
void
FlakFancyVisualizer::setShipStatus(ship_t ship, bool boxed)
{
    if (boxed) {
        if (!ship_info[ship].box_sprite) {
            ship_info[ship].box_sprite = new FlakBoxSprite(*this, ship);
            manager.addSprite(ship_info[ship].box_sprite);
        }
    } else {
        if (ship_info[ship].box_sprite) {
            ship_info[ship].box_sprite->deleteThis();
            ship_info[ship].box_sprite = 0;
        }
    }
}

void
FlakFancyVisualizer::destroyObject(FlakObject& obj, bool violent)
{
    (void) violent;
    setObjectSprite(obj, 0, false);
}

void
FlakFancyVisualizer::destroyShip(ship_t unit)
{
    const FlakShip& sh = getBattle().getShipByNumber(unit);
    getManager().addSprite(new FlakExplSprite(sh.getPos(), *this, 5, 4));
}

void
FlakFancyVisualizer::fireBeam(FlakObject& obj1, FlakObject& obj2)
{
    getManager().addSprite(new FlakBeamSprite(obj1.position,
                                              obj2.position,
                                              *this));
    getManager().addSprite(new FlakExplSprite(obj2.position, *this, 2, 4));
}

void
FlakFancyVisualizer::fireBeam(FlakObject& obj1, ship_t unit2)
{
    getManager().addSprite(new FlakBeamSprite(obj1.position,
                                              getBattle().getShipByNumber(unit2).getPos(),
                                              *this));
}

void
FlakFancyVisualizer::fireBeam(ship_t unit1, FlakObject& obj2, bool hits)
{
    getManager().addSprite(new FlakBeamSprite(getBattle().getShipByNumber(unit1).getPos(),
                                              obj2.position,
                                              *this));
    if (hits)
        getManager().addSprite(new FlakExplSprite(obj2.position, *this, 2, 4));
}

void
FlakFancyVisualizer::fireBeam(ship_t unit1, ship_t unit2, bool hits)
{
    (void) hits;
    getManager().addSprite(new FlakBeamSprite(getBattle().getShipByNumber(unit1).getPos(),
                                              getBattle().getShipByNumber(unit2).getPos(),
                                              *this));
}

void
FlakFancyVisualizer::hitShip(ship_t unit, int expl, int kill, int death_flag)
{
    (void) expl;
    (void) kill;
    (void) death_flag;

    markShipChanged(unit);
    const FlakShip& sh = getBattle().getShipByNumber(unit);
    if (sh.status.shield <= 0) {
        getManager().addSprite(new FlakExplSprite(sh.getPos(), *this, 1 + rand()%4, 4));
    }
}
