/**
  *  \file game/shippredict.cc
  *  \brief Ship Prediction
  *
  *  \todo GShipTurnPredictor operates on TShip objects, not GShip.
  *  This is ugly, as it gives all sorts of problems when requiring
  *  access to ship properties not stored in TShip. However, the
  *  external interface is nice enough, so we don't need to convert
  *  this code until we actually implement the additional
  *  functionality in this module (experience support, per-ship
  *  functions).
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "game/shippredict.h"
#include "game/univ.h"
#include "game/reg.h"
#include "game/host.h"
#include "game/mission.h"
#include "game/config.h"
#include "game/formula.h"
#include "util/recode.h"

const int GShipTurnPredictor::MOVEMENT_TIME_LIMIT;

/** Perform refinery reaction.
    \param s   [in/out] Ship
    \param ore [in/out] Ore to consume
    \param sup [in/out] Supplies to consume */
static void
doRefinery(TShip& s, int16_t& ore, int16_t& sup)
{
    int n = getHull(s.hull).getFuel() - s.ore_cargo[el_Neutronium];
    if (n > ore)
        n = ore;
    if (n > sup)
        n = sup;
    s.ore_cargo[el_Neutronium] += n;
    sup -= n;
    ore -= n;
}

/** Perform direct alchemy reaction (Alchemy + Refinery function on one ship).
    \param ship [in/out] Ship
    \param ratio [in] Supply conversion ratio */
static void
doDirectRefinery(TShip& ship, const int ratio)
{
    int free = getHull(ship.hull).getFuel() - ship.ore_cargo[el_Neutronium];
    int prod = std::min(ship.supplies / ratio, free);
    ship.supplies -= prod*ratio;
    ship.ore_cargo[el_Neutronium] += prod;
}

/** Perform alchemy reaction.
    Merlin converts 3 Supplies -> 1 Mineral.
    Friendly codes allow target mineral control.

    THost and NuHost actually convert 9 Supplies -> 3 Minerals;
    this is what the %rounder is used for. */
static void
doAlchemy(TShip& ship)
{
    // Merlin converts 3 Sup -> 1 Min
    int t = 0, d = 0, m = 0;
    int mins = ship.supplies / 3;
    int rounder = host.isPHost() ? 1 : 3;
    if (allowRegistered(ship.owner)) {
        if (std::memcmp(ship.fcode, "alt", 3) == 0) {
            t = rounder*(mins/rounder);
        } else if (std::memcmp(ship.fcode, "ald", 3) == 0) {
            d = rounder*(mins/rounder);
        } else if (std::memcmp(ship.fcode, "alm", 3) == 0) {
            m = rounder*(mins/rounder);
        } else if (host.isPHost()) {
            if (std::memcmp(ship.fcode, "nat", 3) == 0) {
                d = m = mins/2;
            } else if (std::memcmp(ship.fcode, "nad", 3) == 0) {
                t = m = mins/2;
            } else if (std::memcmp(ship.fcode, "nam", 3) == 0) {
                t = d = mins/2;
            } else {
                t = d = m = mins/3;
            }
        } else {
            t = d = m = mins/3;
        }
    } else {
        t = d = m = mins/3;
    }

    ship.supplies -= 3*(t+d+m);
    ship.ore_cargo[el_Tritanium]  += t;
    ship.ore_cargo[el_Duranium]   += d;
    ship.ore_cargo[el_Molybdenum] += m;
}

static int
getCloakFuel(const TShip& s, int turns, int real_owner)
{
    const int cfb = config.CloakFuelBurn(real_owner);
    int fuel = getHull(s.hull).getMass() * cfb / 100;
    if (fuel < cfb)
        fuel = cfb;
    if (turns)
        fuel *= turns;
    return fuel;
}

static void
normalizePosition(TShip& ship)
{
    GPoint new_pos = GPoint(ship.x, ship.y).getCanonicalLocation();
    ship.x = new_pos.x;
    ship.y = new_pos.y;
}

static int
getEngineLoad(const GUniverse& univ, TShip& ship, TShip* towee_override, bool tow_corr)
{
    int mass = getShipMass(ship);
    if (ship.mission == GMission::msn_Tow && ship.mission_tow > 0 && ship.mission_tow <= NUM_SHIPS) {
        int towee_mass = 0;
        if (towee_override && towee_override->ship_id == ship.mission_tow) {
            towee_mass = getShipMass(*towee_override);
        } else if (univ.ty_any_ships.isValidIndex(ship.mission_tow)) {
            mp16_t t = univ.getShip(ship.mission_tow).getMass();
            if (t.isKnown()) {
                towee_mass = t;
            }
        }
        if (tow_corr)
            towee_mass = 10*(towee_mass/10);
        mass += towee_mass;
    }
    return mass;
}

static int
computeFuelUsage(const GUniverse& univ, TShip& ship, TShip* towee_override,
                 bool grav_acc, double dist)
{
    if (ship.warp <= 0)
        return 0;

    int way = ship.warp*ship.warp;
    if (grav_acc)
        way *= 2;

    int load = getEngineLoad(univ, ship, towee_override, host.isTHost());
    long ff = getEngine(ship.engine).getFuelFactor(ship.warp);

    if (host.isTHost()) {
        // THost formula
        return (ff * (load / 10) * long(dist)) / (10000L * way);
    } else if (!config.UseAccurateFuelModel()) {
        // PHost, standard formula
        return rdivaddl(load, 10, 0) * ff * long(dist) / (10000L * way);
    } else {
        // PHost, "accurate" formula
        return roundToInt(load * (1 - std::exp(-(ff * dist) / (way*100000L))));
    }
}


/*************************** GShipTurnPredictor **************************/

/** Constructor. Create a predictor for a single ship.
    \param univ universe containing the ship
    \param id   ship Id */
GShipTurnPredictor::GShipTurnPredictor(const GUniverse& univ, int id)
    : id(id), towee_override(0),
      univ(univ), movement_fuel_used(0), cloak_fuel_used(0), turns_computed(0)
{
    init();
}

/** Constructor. Create a predictor for a ship towing another ship.
    \param univ   universe containing the ship
    \param id     ship Id
    \param towee_override predictor for towed ship */
GShipTurnPredictor::GShipTurnPredictor(const GUniverse& univ, int id, GShipTurnPredictor& towee_override)
    : id(id), towee_override(&towee_override),
      univ(univ), movement_fuel_used(0), cloak_fuel_used(0), turns_computed(0)
{
    init();
}

void
GShipTurnPredictor::init()
{
    if (univ.getShip(id).hasFullShipData()) {
        univ.getShip(id).getShipData(this->ship);
        ship.owner = univ.getShip(id).getRealOwner();
        valid = true;
    } else {
        valid = false;
    }
}

/** Compute one turn. This will compute all changes for the ship. It
    will also update the towed ship predictor, if any. */
void
GShipTurnPredictor::computeTurn()
{
    // is this actually a predictable ship?
    if (!valid)
        return;

    // where are we?
    int pid                = univ.getPlanetAt(GPoint(ship.x, ship.y));
    const GShip& real_ship = univ.getShip(id);

//     {+++ Mine Laying +++}
//     IF (s.Neutronium<>0) AND (ComputeMineDrop(@s, md)) THEN BEGIN
//       Dec(s.Torps, md.torps);
//     END;

    // Special Missions I (Super Refit, Self Repair, Hiss, Rob) would go here
//   {+++ lfm, mkt, ... +++}
//   IF ((GetStr(s.fc, 3)='mkt')
//       OR ((pconf<>NIL) AND (s.Mission = pconf^.main.ExtMissionsStartAt + pmsn_BuildTorpsFromCargo)))
//    AND (s.TLauncher>0) AND (s.TLaunCnt>0) AND (s.Neutronium>0) THEN
//   BEGIN
//     { mkt }
//     i := s.Tritanium;
//     IF i > s.Duranium THEN i := s.Duranium;
//     IF i > s.Molybdenum THEN i := s.Molybdenum;
//     IF Torps[s.TLauncher].TorpCost>0 THEN BEGIN
//       j := s.Money DIV Torps[s.TLauncher].TorpCost;
//       IF i > j THEN i := j;
//     END;
//     Inc(s.Torps, i);
//     Dec(s.Tritanium, i);
//     Dec(s.Duranium, i);
//     Dec(s.Molybdenum, i);
//     Dec(s.Money, Torps[s.TLauncher].TorpCost * i);
//   END;

    /* Alchemy */
    if (std::memcmp(ship.fcode, "NAL", 3) != 0) {
        if (real_ship.canDoSpecial(hf_MerlinAlchemy)) {
            if (config.AllowAdvancedRefinery() && real_ship.canDoSpecial(hf_AriesRefinery)) {
                doDirectRefinery(ship, 3);
            } else if (real_ship.canDoSpecial(hf_NeutronicRefinery)) {
                doDirectRefinery(ship, 4);
            } else {
                doAlchemy(ship);
            }
        } else if (real_ship.canDoSpecial(hf_NeutronicRefinery)) {
            // Neutronic refinery converts 1 Sup + 1 Min -> 1 Neu
            doRefinery(ship, ship.ore_cargo[el_Tritanium],  ship.supplies);
            doRefinery(ship, ship.ore_cargo[el_Duranium],   ship.supplies);
            doRefinery(ship, ship.ore_cargo[el_Molybdenum], ship.supplies);
        } else if (real_ship.canDoSpecial(hf_AriesRefinery)) {
            // Aries converts 1 Min -> 1 Neu
            int16 max = 0x7FFF;  // we assume that no ship has more cargo
            doRefinery(ship, ship.ore_cargo[el_Tritanium],  max);
            doRefinery(ship, ship.ore_cargo[el_Duranium],   max);
            doRefinery(ship, ship.ore_cargo[el_Molybdenum], max);
        }
    }

    /* Starbase Missions I (Fix, [Recycle, Load Torps]) */
    if (pid > 0) {
        const GPlanet& p = univ.getPlanet(pid);
        if (p.isPlayable(GObject::Playable)
            && p.hasFullBaseData()
            && p.getBaseShipyardOrder() == sya_Fix
            && p.getBaseShipyardId() == id)
        {
            // we're at a base which is fixing us
            ship.damage = 0;
            ship.crew   = getHull(ship.hull).getCrew();
        }
    }

    /* Supply Repair */
    if (ship.damage > 0 && ship.supplies > 0) {
        int max = ship.supplies / 5;
        if (max > ship.damage)
            max = ship.damage;
        ship.damage -= max;
        ship.supplies -= 5*max;
    }

    /* Cloak Fuel Burn */
    if ((real_ship.canDoSpecial(hf_Cloak) || real_ship.canDoSpecial(hf_AdvancedCloak))
        && isMissionCloaking(ship.mission, ship.owner))
    {
        int fuel = real_ship.canDoSpecial(hf_AdvancedCloak) ? 0 : getCloakFuel(ship, 0, real_ship.getRealOwner());
        if (ship.ore_cargo[el_Neutronium] <= fuel
            || (ship.damage >= config.DamageLevelForCloakFail()
                && !real_ship.canDoSpecial(hf_HardenedCloak)))
        {
            /* We cancel only cloak missions here. Other missions are
               NOT canceled, see below. */
            ship.mission = 0;
        } else {
            ship.ore_cargo[el_Neutronium] -= fuel;
            cloak_fuel_used += fuel;
        }
    }

    /* This used to check for fuelless ships at this place, and set
       them to mission zero. However, since this is mainly used for
       fuel predictions, this is not a good idea. We want to know how
       much we need, even if you have too little. This mainly affects
       the Tow mission. */

    /* damage speed limit */
    if (ship.damage > 0 && !real_ship.canDoSpecial(hf_HardenedEngines)) {
        int limit = (raceId(ship.owner) == 2
                     ? (host.isPHost()
                        ? 15 - ship.damage/10
                        : 14 - ship.damage/10)
                     : 10 - ship.damage/10);
        if (limit < 0)
            limit = 0;
        if (ship.warp > limit)
            ship.warp = limit;
    }

    /* Actual movement here */
    long dist2 = long(ship.waypoint_dx*ship.waypoint_dx) + long(ship.waypoint_dy*ship.waypoint_dy);
    if (real_ship.canDoSpecial(hf_Hyperdrive)
        && std::memcmp(ship.fcode, "HYP", 3) == 0
        && ship.warp > 0
        && dist2 >= 20*20)   // FIXME: minimum distance not in PHost?
    {
        /* It's hyperjumping */
        ship.ore_cargo[el_Neutronium] -= 50;
        movement_fuel_used            += 50;

        /* If it's jumping, it can't tow. Advance time in towee's
           world anyway. */
        if (ship.mission == GMission::msn_Tow)
            ship.mission = 0;
        if (towee_override)
            towee_override->computeTurn();

        /* Now move that bugger. */
        if (host.isExactHyperjumpDistance2(dist2)) {
            ship.x += ship.waypoint_dx;
            ship.y += ship.waypoint_dy;
        } else {
            // non-exact jump
            double dist = 350 / std::sqrt(double(dist2));
            int mx = int(dist * std::abs(ship.waypoint_dx) + 0.4999999);
            int my = int(dist * std::abs(ship.waypoint_dy) + 0.4999999);
            if (ship.waypoint_dx < 0)
                mx = -mx;
            if (ship.waypoint_dy < 0)
                my = -my;
            ship.x += mx;
            ship.y += my;
        }
        ship.waypoint_dx = ship.waypoint_dy = 0;
        ship.warp = 0;
        normalizePosition(ship);
        // FIXME: gravity wells?
    } else if (dist2 && ship.warp > 0) {
        /* Normal movement */
        // First, compute new position in mx,my
        double dist = std::sqrt(double(dist2));
        int way = ship.warp * ship.warp;
        if (real_ship.canDoSpecial(hf_Gravitonic))
            way *= 2;

        int mx = ship.waypoint_dx, my = ship.waypoint_dy;
        if (dist > way) {
            int dx, dy;
            if (host.isTHost()) {
                // THost movement formulas, from Donovan's
                if (std::abs(mx) > std::abs(my)) {
                    dx = int(double(way) * std::abs(mx) / dist + 0.5);
                    dy = int(double(dx) * std::abs(my) / std::abs(mx) + 0.5);
                } else {
                    dy = int(double(way) * std::abs(my) / dist + 0.5);
                    dx = int(double(dy) * std::abs(mx) / std::abs(my) + 0.5);
                }
                if (mx < 0)
                    dx = -dx;
                if (my < 0)
                    dy = -dy;
            } else {
                // PHost. From docs and source.
                double head = getHeadingRad(mx, my);
                double fx = std::sin(head) * way;
                double fy = std::cos(head) * way;
                dx = int(fx);
                dy = int(fy);
                if (dx != fx)
                    dx += sgn(fx);
                if (dy != fy)
                    dy += sgn(fy);
                if (mx == 0)
                    dx = 0;
                if (my == 0)
                    dy = 0;
            }

            // we now have the dx,dy we want to move
            mx = dx;
            my = dy;
            dist = way;         // FIXME: use distFromDX(dx,dy) instead?
        }

        // Advance time in towee. Must be here because we need to know its "post-movement" mass.
        if (towee_override) {
            if (ship.mission == GMission::msn_Tow) {
                // we assume the tow succeeds. FIXME: be more clever?
                if (towee_override->ship.mission == GMission::msn_Tow)
                    towee_override->ship.mission = 0;
                towee_override->ship.warp = 0;
            }
            towee_override->computeTurn();
        }

        // Compute fuel usage.
        int fuel = computeFuelUsage(univ, ship,
                                    towee_override ? &towee_override->ship : 0,
                                    real_ship.canDoSpecial(hf_Gravitonic),
                                    dist);
        ship.ore_cargo[el_Neutronium] -= fuel;
        movement_fuel_used += fuel;

        // We still have the position offset in mx,my. Move it.
        ship.x += mx;
        ship.y += my;
        ship.waypoint_dx -= mx;
        ship.waypoint_dy -= my;
        normalizePosition(ship);

        // warp wells suck. (the ship to the planet, of course.)
        int wp_x = ship.x + ship.waypoint_dx;
        int wp_y = ship.y + ship.waypoint_dy;

        if (ship.warp > 1 && !univ.getPlanetAt(GPoint(ship.x, ship.y))) {
            if (int gpid = univ.getGravityPlanetAt(GPoint(ship.x, ship.y))) {
                // okay, there is a planet. Move the ship.
                GPoint ppos = univ.getPlanet(gpid).getPos();
                ship.x = ppos.x;
                ship.y = ppos.y;
                // Now adjust the waypoint. If the waypoint was inside
                // the warp well of the same planet, assume the end of
                // this movement order. Otherwise, when users set a
                // waypoint at the edge of the warp well, the ship
                // would try to get there for ever. This is consistent
                // with the CCScript `MoveTo' command and with what
                // most people expect.
                if (univ.getPlanetAt(GPoint(wp_x, wp_y), true) == gpid) {
                    ship.waypoint_dx = ship.waypoint_dy = 0;
                } else {
                    GPoint new_wp = GPoint(wp_x, wp_y).getSimpleNearestAlias(GPoint(ship.x, ship.y));
                    ship.waypoint_dx = new_wp.x - ship.x;
                    ship.waypoint_dy = new_wp.y - ship.y;
                }
            }
        }

        // Update towee position
        if (towee_override) {
            if (ship.mission == GMission::msn_Tow) {
                towee_override->ship.x = ship.x;
                towee_override->ship.y = ship.y;
                towee_override->ship.waypoint_dx = 0;
                towee_override->ship.waypoint_dy = 0;
            }
        }
    } else {
        /* No sensible movement order for this ship. Advance towee's
           time anyway. */
        if (towee_override)
            towee_override->computeTurn();
    }

    /* turn is over. */
    ++turns_computed;
}

/** Compute this ship's movement, until movement is over or time runs out. */
void
GShipTurnPredictor::computeMovement()
{
    if (valid) {
        const int final_turn = turns_computed + MOVEMENT_TIME_LIMIT;
        while ((ship.waypoint_dx || ship.waypoint_dy) && turns_computed < final_turn) {
            computeTurn();
            if (ship.ore_cargo[el_Neutronium] < 0)
                ship.ore_cargo[el_Neutronium] = 0;
        }
    }
}

/** Override this ship's position. */
void
GShipTurnPredictor::setPosition(GPoint pt)
{
    pt = pt.getSimpleCanonicalLocation();
    ship.x = pt.x;
    ship.y = pt.y;
}

/** Override this ship's waypoint. */
void
GShipTurnPredictor::setWaypoint(GPoint pt)
{
    pt = pt.getSimpleNearestAlias(GPoint(ship.x, ship.y));
    ship.waypoint_dx = pt.x - ship.x;
    ship.waypoint_dy = pt.y - ship.y;
}

/** Override this ship's friendly code. */
void
GShipTurnPredictor::setFCode(string_t s)
{
    storeBasicString(ship.fcode, convertUtf8ToGame(s));
}

/** Get this ship's friendly code. */
string_t
GShipTurnPredictor::getFCode() const
{
    return convertGameToUtf8(getBasicString(ship.fcode));
}
