/**
  *  \file int/notify.cc
  *  \brief Notification Messages
  *
  *  Notification messages provide a way for background scripts (in
  *  particular, auto tasks) to inform the user. The intent is to
  *  provide some kind of "push" interface, actively presenting these
  *  messages to the users, unlike the easy-to-oversee interface like
  *  the console.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/notify.h"
#include "int/exec.h"
#include "util/nls.h"


/** Constructor. To be called by IntNotificationMessageStore::addNewMessage. */
IntNotificationMessage::IntNotificationMessage(string_t text, IntNotificationMessageStore& store)
    : text(text),
      confirmed(false),
      associated_process(0),
      store(store)
{ }

IntNotificationMessage::~IntNotificationMessage()
{
    // Deletion of a message dissociates it from its process
    if (associated_process != 0) {
        associated_process->setNotificationMessage(0);
    }
}

/** Set this message's text. */
void
IntNotificationMessage::setText(string_t text)
{
    if (text != this->text) {
        this->text = text;
        store.sig_changed.raise();
    }
}

/** Get this message's text. */
string_t
IntNotificationMessage::getText() const
{
    return text;
}

/** Set confirmation status of this message. */
void
IntNotificationMessage::setConfirmed(bool value)
{
    if (value != this->confirmed) {
        this->confirmed = value;
        store.sig_changed.raise();
    }
}

/** Remove this message. This deletes (invalidates) the
    IntNotificationMessage object. */
void
IntNotificationMessage::remove()
{
    store.removeMessageByIndex(store.getIndexOfMessage(this));
}

/********************** IntNotificationMessageStore **********************/

/** Index indicating "none/not found". */
const IntNotificationMessageStore::index_t IntNotificationMessageStore::nil;

/** Global notification message instance. */
IntNotificationMessageStore IntNotificationMessageStore::instance;

/** Constructor. */
IntNotificationMessageStore::IntNotificationMessageStore()
{ }

/** Destructor. */
IntNotificationMessageStore::~IntNotificationMessageStore()
{ }

/** Get number of messages in the message store. */
IntNotificationMessageStore::index_t
IntNotificationMessageStore::getNumMessages() const
{
    return messages.size();
}

/** Get index of a message, given the IntNotificationMessage object.
    \return index, nil if message not contained in this store. */
IntNotificationMessageStore::index_t
IntNotificationMessageStore::getIndexOfMessage(IntNotificationMessage* msg)
{
    for (index_t i = 0; i < messages.size(); ++i)
        if (messages[i] == msg)
            return i;
    return nil;
}

/** Get message by index.
    \param index Message number, [0,getNumMessages()) */
IntNotificationMessage*
IntNotificationMessageStore::getMessageByIndex(index_t index)
{
    return messages[index];
}

/** Add new message.
    \param text Text of this message */
IntNotificationMessage*
IntNotificationMessageStore::addNewMessage(string_t text)
{
    IntNotificationMessage* nm = new IntNotificationMessage(text, *this);
    messages.push_back(nm);
    sig_changed.raise();
    return nm;
}

/** Remove message, given an index. */
void
IntNotificationMessageStore::removeMessageByIndex(index_t idx)
{
    messages.erase(messages.begin() + idx);
    if (idx < getCurrent()) {
        setCurrent(getCurrent()-1);
    }
    sig_changed.raise();
}

/** Make all confirmed processes runnable. Only affects processes that are
    actually suspended. In particular, frozen processes are not changed. */
void
IntNotificationMessageStore::makeConfirmedProcessesRunnable()
{
    for (index_t i = 0; i < messages.size(); ++i) {
        IntExecutionContext* p = messages[i]->getAssociatedProcess();
        if (p != 0
            && messages[i]->isConfirmed()
            && p->getState() == IntExecutionContext::psSuspended)
        {
            p->setState(IntExecutionContext::psRunnable);
        }
    }
}

int
IntNotificationMessageStore::getCount()
{
    return getNumMessages();
}

string_t
IntNotificationMessageStore::getText(int index)
{
    IntNotificationMessage* msg = messages[index];

    string_t text = msg->getText();
    if (msg->getAssociatedProcess() != 0) {
        if (msg->isConfirmed()) {
            text += _("\n\nThis message has been confirmed.");
        } else if (msg->getAssociatedProcess()->getProcessKind() != IntExecutionContext::pkDefault) {
            text += _("\n\nThe auto task has been stopped; it will continue when you confirm this message.");
        } else {
            text += _("\n\nThe script has been stopped; it will continue when you confirm this message.");
        }
    }
    return text;
}

string_t
IntNotificationMessageStore::getHeading(int index)
{
    string_t head = messages[index]->getText();

    // Remove everything after first line
    string_t::size_type n = head.find('\n');
    if (n != string_t::npos) {
        head.erase(n);
    }

    // If message starts with '(-sXXX)', remove XXX
    if (head.size() > 3 && head[0] == '(') {
        head.erase(3, head.find(')', 3) - 3);
    }

    // Remove all angle brackets
    while ((n = head.find_first_of("<>")) != string_t::npos) {
        head.erase(n, 1);
    }

    return head;
}
