/**
  *  \file ui/widgets/inputline.cc
  *  \brief Input-line Widget
  *
  *  \todo This is currently not too efficient because it parses UTF-8
  *  a lot.
  *
  *  (c) 2001-2013 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "ui/widgets/inputline.h"
#include "cpluslib/utf8.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/util.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/nls.h"
#include "util/recode.h"
#include "ui/widgets/button.h"
#include "ui/uiglobal.h"


static string_t::size_type
getPreviousWordBoundary(const string_t& str, string_t::size_type pos)
{
    while (pos > 0 && charAtUtf8(str, pos-1) == ' ')
        --pos;
    while (pos > 0 && charAtUtf8(str, pos-1) != ' ')
        --pos;
    return pos;
}

static string_t::size_type
getNextWordBoundary(const string_t& str, string_t::size_type pos)
{
    const string_t::size_type max = lengthUtf8(str);
    while (pos < max && charAtUtf8(str, pos) == ' ')
        ++pos;
    while (pos < max && charAtUtf8(str, pos) != ' ')
        ++pos;
    return pos;
}

/****************************** UIInputLine ******************************/

/** \class UIInputLine
    \brief Input Line

    Standard text entry field */

/** Input line.
    \param amaxl   maximum length of input (a hard limit)
    \param ahot    hot-key that brings focus to this widget
    \param aflags  options
    \param aid     widget Id */
UIInputLine::UIInputLine(string_t::size_type amaxl, UIKey ahot,
                         int aflags, int aid)
    : UIWidget(aid), hotkey(ahot), maxlen(amaxl), curx(0), firstx(0),
      text(), font(FONT_TITLE), pref_len(amaxl < 40 ? amaxl : 40)
{
    setState(aflags | ilf_TypeErase, true);
}

/** Input line.
    \param amaxl   maximum length of input (a hard limit)
    \param apref_len preferred width of input, in "em" widths
    \param ahot    hot-key that brings focus to this widget
    \param aflags  options
    \param aid     widget Id */
UIInputLine::UIInputLine(string_t::size_type amaxl, int apref_len,
                         UIKey ahot,
                         int aflags, int aid)
    : UIWidget(aid), hotkey(ahot), maxlen(amaxl), curx(0), firstx(0),
      text(), font(FONT_TITLE), pref_len(apref_len)
{
    setState(aflags | ilf_TypeErase, true);
}

/** Set contents of input line. */
void
UIInputLine::setText(string_t s)
{
    text = s;
    curx = lengthUtf8(text);
    firstx = 0;
    scroll();
    drawWidget(false);
    sig_changed.raise(getId());
}

void
UIInputLine::onStateChange(int astate, bool enable)
{
    UIWidget::onStateChange(astate, enable);
    if ((astate & st_Focused) && enable)
        status |= ilf_TypeErase;
    if (astate & (st_Focused | st_Disabled | ilf_TypeErase))
        drawWidget(false);
}

/** Get width of space available for text. */
int
UIInputLine::getTextWidth()
{
    int rv = extent.w;
    if (hasState(ilf_Framed))
        rv -= 4;
    return rv;
}

/** Get height of space needed for text. */
int
UIInputLine::getTextHeight()
{
    int rv = fonts[font]->getTextHeight("Tp");
    if (hasState(ilf_Framed))
        return rv + 4;
    else
        return rv;
}

/** Check whether an Unicode character should be accepted. */
bool
UIInputLine::acceptUnicode(uint32_t uni) const
{
    if (uni < 32 || uni == 127) {
        // Refuse controls
        return false;
    }
    if (hasState(ilf_NumbersOnly) && (uni < '0' || uni > '9')) {
        // Refuse nondigits if requested
        return false;
    }
    if (hasState(ilf_NoHi) && (uni >= 128)) {
        // Refuse non-ASCII if requested
        return false;
    }
    if (hasState(ilf_GameChars) && !isReliableGameCharacter(uni)) {
        // Refuse non-game characters if requested
        return false;
    }
    return true;
}

/** Adjust display so that cursor is visible. */
void
UIInputLine::scroll()
{
    const string_t::size_type textLength = lengthUtf8(text);

    if (curx > textLength)
        curx = textLength;
    if (curx < 3)
        firstx = 0;
    else if (firstx > curx - 3)
        firstx = curx - 3;

    string_t t = hasState(ilf_Hidden) ? string_t(textLength, '*') : text;
    int ln = fonts[font]->getTextWidth(substrUtf8(t, firstx, curx - firstx));
    int av = getTextWidth() - CursorWidth;
    if (curx) {
        while (firstx < curx-1 && ln > av) {
            ++firstx;
            ln = fonts[font]->getTextWidth(substrUtf8(t, firstx, curx - firstx));
        }
    }
}

/** Insert text at current cursor position. Respects flags
    ilf_NonEditable (=request is ignored) and ilf_TypeErase (=replace
    current text). */
void
UIInputLine::insert(string_t s)
{
    if (hasState(ilf_NonEditable))
        return;
    if (hasState(ilf_TypeErase)) {
        text = s;
        curx = lengthUtf8(s);
        setState(ilf_TypeErase, false);
    } else {
        // Can we insert?
        string_t::size_type textLength = lengthUtf8(text);
        if (textLength >= maxlen)
            return;

        // Limit inserted text
        s = substrUtf8(s, 0, maxlen - textLength);

        // Compute new text
        text = substrUtf8(text, 0, curx) + s + substrUtf8(text, curx, string_t::npos);
        curx += lengthUtf8(s);
    }
    scroll();
    drawWidget(false);
    sig_changed.raise(getId());
}

/** Move cursor to position pos.
    \param pos cursor position, [0, lengthUtf8()] */
void
UIInputLine::setCursor(pos_type pos)
{
    const string_t::size_type textLength = lengthUtf8(text);
    if (pos > textLength)
        pos = textLength;
    if (pos != curx) {
        setState(ilf_TypeErase, false);
        curx = pos;
        scroll();
        drawWidget(false);
    }
}

/** Get cursor position. Note that this position is in runes, not bytes! */
UIInputLine::pos_type
UIInputLine::getCursor() const
{
    return curx;
}

bool
UIInputLine::handleEvent(const UIEvent& e, bool second_pass)
{
    switch(e.type) {
     case ev_Mouse:
        if (e.mouse.b && isInside(GfxPoint(e.mouse.x, e.mouse.y)) && !hasState(st_Disabled)) {
            setSelectedWidget();
            setFocusedWidget();

            pos_type new_pos = firstx;
            int      x0      = hasState(ilf_Framed) ? extent.x + 2 : extent.x;
            if (e.mouse.x < x0) {
                if (new_pos)
                    --new_pos;
            } else {
                int delta = e.mouse.x - x0;
                if (hasState(ilf_Hidden)) {
                    while (new_pos < lengthUtf8(text) &&
                           fonts[font]->getTextWidth(string_t(new_pos - firstx, '*')) < delta)
                        ++new_pos;
                } else {
                    while (new_pos < lengthUtf8(text) &&
                           fonts[font]->getTextWidth(substrUtf8(text, firstx, new_pos - firstx)) < delta)
                        ++new_pos;
                }
            }
            // FIXME: consumeprefix()
            curx = new_pos;
            drawWidget(false);
            return true;
        }
        break;
     case ev_Key:
        if (hasState(st_Disabled)) {
            break;
        } else if (!hasState(st_Focused)) {
            if (e.key.code == hotkey) {
                setSelectedWidget();
                setFocusedWidget();
                return true;
            }
        } else {
            string_t::size_type pos;
            switch(e.key.code) {
             case SDLK_LEFT:
                /* Move cursor left */
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                if (curx > 0)
                    setCursor(curx - 1);
                return true;
             case SDLK_RIGHT:
                /* Move cursor right */
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                if (curx < lengthUtf8(text))
                    setCursor(curx + 1);
                return true;
             case ss_Ctrl + SDLK_LEFT:
                /* Move cursor left one word */
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                setCursor(getPreviousWordBoundary(text, curx));
                return true;
             case ss_Ctrl + SDLK_RIGHT:
                /* Move cursor right one word */
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                setCursor(getNextWordBoundary(text, curx));
                return true;
             case SDLK_BACKSPACE:
                /* Delete character backward */
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                if (curx > 0) {
                    --curx;
                    if (!hasState(ilf_NonEditable)) {
                        text = substrUtf8(text, 0, curx) + substrUtf8(text, curx+1, string_t::npos);
                        sig_changed.raise(getId());
                    }
                    drawWidget(false);
                }
                return true;
             case SDLK_DELETE:
             case ss_Ctrl + 'd':
                /* Delete character forward, or delete whole text */
                /* Note that in PCC 1.x, Ctrl-D is delete word forward */
                setSelectedWidget();
                if (!hasState(ilf_NonEditable)) {
                    if (hasState(ilf_TypeErase)) {
                        text.erase();
                        setState(ilf_TypeErase, false);
                    } else {
                        if (curx < lengthUtf8(text)) {
                            text = substrUtf8(text, 0, curx) + substrUtf8(text, curx+1, string_t::npos);
                        }
                    }
                    drawWidget(false);
                    sig_changed.raise(getId());
                }
                return true;
             case SDLK_HOME:
                /* Go to beginning */
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                setCursor(0);
                return true;
             case SDLK_END:
                /* Go to end */
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                setCursor(lengthUtf8(text));
                return true;
             case ss_Ctrl + 'y':
                /* Delete whole line */
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                if (!hasState(ilf_NonEditable)) {
                    text.erase();
                    drawWidget(false);
                    sig_changed.raise(getId());
                }
                setCursor(0);
                return true;
             case ss_Ctrl + 'k':
                /* Delete till end of line */
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                if (!hasState(ilf_NonEditable)) {
                    text = substrUtf8(text, 0, curx);
                    drawWidget(false);
                    sig_changed.raise(getId());
                }
                return true;
             case ss_Ctrl + SDLK_BACKSPACE:
                /* Delete word backward */
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                if (!hasState(ilf_NonEditable)) {
                    pos = getPreviousWordBoundary(text, curx);
                    text = substrUtf8(text, 0, pos) + substrUtf8(text, curx, string_t::npos);
                    setCursor(pos);
                    drawWidget(false);
                    sig_changed.raise(getId());
                }
                return true;
             case ss_Ctrl + 'u':
                /* No-op, just clear type-erase */
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                return true;
             case ss_Ctrl + 't': {
                /* Transpose characters */
                string_t::size_type n = lengthUtf8(text);
                setSelectedWidget();
                setState(ilf_TypeErase, false);
                if (n >= 2 && !hasState(ilf_NonEditable)) {
                    pos = curx;
                    if (pos >= n)
                        pos = n - 1;
                    text = substrUtf8(text, 0, pos-1) + substrUtf8(text, pos, 1) + substrUtf8(text, pos-1, 1) + substrUtf8(text, pos+1, string_t::npos);
                    setCursor(pos+1);
                    drawWidget(false);
                    sig_changed.raise(getId());
                }
                return true;
             }

             default:
                if ((e.key.code & ss_AllMod) == 0 && acceptUnicode(e.key.unicode)) {
                    /* Self-insert */
                    string_t n;
                    appendUtf8(n, e.key.unicode);
                    setSelectedWidget();
                    insert(n);
                    return true;
                }
                break;
            }
        }
        break;
     default:
        break;
    }
    return UIWidget::handleEvent(e, second_pass);
}

void
UIInputLine::drawContent(GfxCanvas& can)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    GfxRect r = getExtent();
    if (hasState(ilf_Framed)) {
        drawFrameDown(can, r);
        r.grow(-2, -2);
    }
    int bgcolor;
    if (getFocusState() != NoFocus) {
        scroll();
        if (hasState(ilf_TypeErase)) {
            ctx.setColor(COLOR_GRAY);
            bgcolor = COLOR_BLUE;
        } else {
            ctx.setColor(COLOR_BLUE);
            bgcolor = COLOR_GRAY;
        }
    } else {
        if (hasState(st_Disabled)) {
            ctx.setColor(COLOR_DARK);
        } else {
            ctx.setColor(COLOR_BLACK);
        }
        bgcolor = COLOR_GRAY;
        firstx = 0;
    }

    string_t t = (hasState(ilf_Hidden)
                  ? string_t(lengthUtf8(text) - firstx, '*')
                  : string_t(substrUtf8(text, firstx, string_t::npos)));
    ctx.useFont(*fonts[font]);
    drawSolidBar(ctx, r, bgcolor);
    outTextF(ctx, r.x, r.y, r.w, t);
    /* outTextF sets cursor to the end of the output */
    drawSolidBar(can, GfxRect(ctx.getCursor().x, r.y, r.x + r.w - ctx.getCursor().x, r.h), standard_colors[COLOR_GRAY]);
    if (getFocusState() != NoFocus && curx >= firstx) {
        unsigned cx = fonts[font]->getTextWidth(substrUtf8(t, 0, curx-firstx));
        if (cx + 10 < r.w)
            /* FIXME: magic numbers */
            drawSolidBar(can, GfxRect(r.x + cx, r.y + 19, 10, 2), standard_colors[COLOR_BLACK]);
    }
    can.update(getExtent());
}

void
UIInputLine::getLayoutInfo(LayoutInfo& info)
{
    info.flags     = li_FlexibleH;
    info.min_size  = GfxPoint(4 * font_ems[font], getTextHeight());
    info.pref_size = GfxPoint(pref_len * font_ems[font], getTextHeight());
}

/** Do a standard input line dialog.
    \param str     string to enter/modify
    \param maxlen  maximum length for string
    \param title,req  title and prompt of dialog window
    \param width   width of input field in "em" widths
    \param flags   flags for input field */
bool
UIInputLine::run(string_t& str, string_t::size_type maxlen,
                 const string_t& title, const string_t& req,
                 int width, int flags)
{
    if (width * font_ems[FONT_TITLE] > ui_root->getExtent().w - 40)
        width = (ui_root->getExtent().w - 40) / font_ems[FONT_TITLE];

    UIWindow win(title, UIWindow::BLUE_WINDOW, COLOR_GRAY, 10);
    win.setLayout(UIVBoxLayout::instance5);

    UIStaticText w_text(req, UIStaticText::StaticTitle);
    win.add(w_text);

    UIInputLine il(maxlen, width, 0, flags, 0);
    win.add(il);

    UIGroup buttons(UIHBoxLayout::instance5);
    UISpacer button_spacer;
    UIButton button_ok(1, _("OK"), SDLK_RETURN);
    UIButton button_cancel(0, _("Cancel"), SDLK_ESCAPE);
    buttons.add(button_spacer);
    buttons.add(button_ok);
    buttons.add(button_cancel);
    button_ok.addStop();
    button_cancel.addStop();
    win.add(buttons);

    win.pack();
    win.setFocusedWidget(il);
    il.setText(str);
    ui_root->centerWidget(win);
    ui_root->add(win);
    int rc = ui_root->run();
    ui_root->close(win);
    if (rc)
        str = il.getText();
    return rc;
}
